@@ -16,30 +16,29 @@ import abc
 import functools
 import itertools
 import logging
+from collections.abc import Generator, Iterable
 from copy import deepcopy
 from dataclasses import dataclass, field
 from typing import Any
-from collections.abc import Generator, Iterable
 
 import networkx as nx
 import pandas as pd
 
-from mostlyai.sdk._data.exceptions import MostlyDataException
-from mostlyai.sdk.domain import ModelEncodingType
-
 from mostlyai.sdk._data.dtype import V_DTYPE_ENCODING_TYPE_MAP, DType, VirtualDType
+from mostlyai.sdk._data.exceptions import MostlyDataException
 from mostlyai.sdk._data.util.common import (
+    DATA_TABLE_METADATA_FIELDS,
+    IS_NULL,
+    MAX_SCP_SIBLINGS_LIMIT,
+    NON_CONTEXT_COLUMN_INFIX,
+    TABLE_COLUMN_INFIX,
+    Key,
+    OrderBy,
     as_list,
     decrypt,
     get_passphrase,
-    MAX_SCP_SIBLINGS_LIMIT,
-    TABLE_COLUMN_INFIX,
-    NON_CONTEXT_COLUMN_INFIX,
-    IS_NULL,
-    DATA_TABLE_METADATA_FIELDS,
-    Key,
-    OrderBy,
 )
+from mostlyai.sdk.domain import ModelEncodingType
 
 _LOG = logging.getLogger(__name__)
 
@@ -189,13 +188,17 @@ class Schema:
             return ctx_rel.child
 
     def get_primary_key(self, table_name: str) -> DataIdentifier | None:
-        context_relations = self.get_child_context_relations(table_name)
-        # first, check primary key on table
+        primary_key = None
         if self.tables[table_name].primary_key is not None:
-            return DataIdentifier(table_name, self.tables[table_name].primary_key)
-        # fall back to checking primary key on schema
-        if context_relations:
-            return context_relations[0].parent
+            # first, check primary key on table
+            primary_key = DataIdentifier(table_name, self.tables[table_name].primary_key)
+        elif context_relations := self.get_child_context_relations(table_name):
+            # fall back to checking primary key on schema
+            primary_key = context_relations[0].parent
+        # lastly, check if the primary key is in the included columns
+        if primary_key and primary_key.column not in self.tables[table_name].columns:
+            primary_key = None
+        return primary_key
 
     def get_child_context_relations(self, parent_table: str) -> list[DataRelation]:
         return [rel for rel in self.relations if rel.parent.table == parent_table and isinstance(rel, ContextRelation)]
@@ -255,7 +258,7 @@ class Schema:
         relations = [rel for rel in relations if isinstance(rel, ContextRelation)]
         return relations
 
-    def update_key_encoding_types(self) -> None:
+    def _update_key_encoding_types(self) -> None:
         for tbl_name, tbl_table in self.tables.items():
             if tbl_table.primary_key is not None:
                 if tbl_table.primary_key in tbl_table.encoding_types:
@@ -265,7 +268,7 @@ class Schema:
                     del tbl_table.encoding_types[rel.child.column]
             self.tables[tbl_name] = tbl_table
 
-    def resolve_auto_encoding_types(self) -> None:
+    def _resolve_auto_encoding_types(self) -> None:
         for tbl_name, tbl_table in self.tables.items():
             for col_name, encoding_type in tbl_table.encoding_types.items():
                 if encoding_type == ModelEncodingType.auto:
@@ -276,7 +279,7 @@ class Schema:
                         f"{promoted_enctype.value} for {tbl_name}.{col_name}"
                     )
 
-    def remove_cascading_keys_relations(self):
+    def _remove_cascading_keys_relations(self):
         tables_with_cascading_keys = set()
         for table_name, table in self.tables.items():
             primary_key = table.primary_key
@@ -294,6 +297,11 @@ class Schema:
         if tables_with_cascading_keys:
             _LOG.info(f"Removed cascading keys relations for tables: {tables_with_cascading_keys}")
 
+    def preprocess_schema_before_pull(self):
+        self._remove_cascading_keys_relations()
+        self._update_key_encoding_types()
+        self._resolve_auto_encoding_types()
+
     @property
     @functools.cache
     def graph(self):
@@ -442,6 +450,16 @@ class DataContainer(abc.ABC):
         """
         pass
 
+    @abc.abstractmethod
+    def query(self, sql: str) -> pd.DataFrame:
+        """
+        Execute a SQL query against this data source.
+
+        :param sql: SQL query to execute
+        :return: DataFrame containing the query results
+        """
+        pass
+
     # DEFAULT METHODS (noop, unless applicable)
     def drop_all(self):
         """
@@ -534,9 +552,9 @@ class DataTable(abc.ABC):
     def read_data(
         self,
         where: dict[str, Any] | None = None,
-        limit_n_rows: int | None = None,
+        limit: int | None = None,
         columns: list[str] | None = None,
-        is_shuffle: bool | None = False,
+        shuffle: bool | None = False,
         order_by: OrderBy | None = None,
         do_coerce_dtypes: bool | None = False,
     ) -> pd.DataFrame:
@@ -544,9 +562,9 @@ class DataTable(abc.ABC):
         Read data from this data source.
 
         :param where: dict of column name and value(s) or a lambda function to filter on
-        :param limit_n_rows: limit the number of rows returned. If None, return all rows
+        :param limit: limit the number of rows returned. If None, return all rows
         :param columns: list of columns to include. If None, all columns are included
-        :param is_shuffle: whether to shuffle the returned data rows
+        :param shuffle: whether to shuffle the returned data rows
         :param order_by: optionally, one or multiple columns (asc/desc) to order by
         :param do_coerce_dtypes: bool on whether dtypes shall be converted corresponding to encoding_types
         """
