@@ -14,40 +14,45 @@
 
 import abc
 import base64
+import decimal
+import enum
 import functools
 import hashlib
 import logging
+import os
 import re
 import shutil
-import os
-
 import socket
 import subprocess
 import tempfile
 import time
+import uuid
 from collections import defaultdict
+from collections.abc import Callable, Generator, Iterable
 from contextlib import contextmanager
 from functools import lru_cache
 from pathlib import Path
 from typing import Any, Literal, Optional
-from collections.abc import Callable, Generator, Iterable
 
+import numpy as np
 import pandas as pd
 import sqlalchemy as sa
 import sqlalchemy.sql.sqltypes as sa_types
 import sshtunnel
 from joblib import Parallel, delayed
+from sqlalchemy import Table
 from sqlalchemy.orm import sessionmaker
 
-from mostlyai.sdk._data.exceptions import MostlyDataException
 from mostlyai.sdk._data.base import (
     DataContainer,
     DataTable,
-    order_df_by,
     ForeignKey,
+    order_df_by,
 )
 from mostlyai.sdk._data.db.types_coercion import coerce_to_sql_dtype
 from mostlyai.sdk._data.dtype import (
+    FLOAT64,
+    STRING,
     VirtualBoolean,
     VirtualDate,
     VirtualDatetime,
@@ -59,9 +64,9 @@ from mostlyai.sdk._data.dtype import (
     WrappedDType,
     coerce_dtypes_by_encoding,
 )
-from mostlyai.sdk._data.util.common import prepare_ssl_path, ColumnSort, OrderBy
+from mostlyai.sdk._data.exceptions import MostlyDataException
+from mostlyai.sdk._data.util.common import ColumnSort, OrderBy, assert_read_only_sql, prepare_ssl_path
 from mostlyai.sdk._data.util.kerberos import is_kerberos_ticket_alive
-from sqlalchemy import Table
 
 _LOG = logging.getLogger(__name__)
 
@@ -587,6 +592,15 @@ class SqlAlchemyContainer(DBContainer, abc.ABC):
     def does_database_exist(self) -> bool:
         pass
 
+    def query(self, sql: str) -> pd.DataFrame:
+        assert_read_only_sql(sql)
+        with self.init_sa_connection() as engine:
+            try:
+                return pd.read_sql(sql, engine, dtype_backend="pyarrow")
+            except sa.exc.SQLAlchemyError as e:
+                _LOG.error(f"Error executing query: {str(e)}")
+                raise
+
     def drop_all(self):
         if self.sa_metadata:
             with self.use_sa_engine() as sa_engine:
@@ -725,6 +739,7 @@ class SqlAlchemyContainer(DBContainer, abc.ABC):
             self.update_dbschema(None)
             return sorted(list(set(self.all_schemas())))
         else:
+            prefix = prefix.rstrip(".")  # remove trailing dot if present
             # List the tables of the provided schema named {prefix}
             self.update_dbschema(prefix)
             locations = [".".join([prefix, loc]) for loc in self.get_object_list()]
@@ -1191,9 +1206,9 @@ class SqlAlchemyTable(DBTable, abc.ABC):
     def read_data(
         self,
         where: dict[str, Any] | None = None,
-        limit_n_rows: int | None = None,
+        limit: int | None = None,
         columns: list[str] | None = None,
-        is_shuffle: bool | None = None,
+        shuffle: bool | None = None,
         order_by: OrderBy | None = None,
         do_coerce_dtypes: bool | None = False,
     ) -> pd.DataFrame:
@@ -1205,15 +1220,15 @@ class SqlAlchemyTable(DBTable, abc.ABC):
         if len(stmts) == 1 and self.ENABLE_ORDER_AND_LIMIT_ON_SQL:
             # single statement; include ordering and limiting in the SQL statement
             # and execute it
-            stmt = self._sa_order(stmts[0], is_shuffle, order_by)
-            stmt = self._sa_limit(stmt, limit_n_rows)
+            stmt = self._sa_order(stmts[0], shuffle, order_by)
+            stmt = self._sa_limit(stmt, limit)
             df = self._sa_execute([stmt])
         else:  # len(stmts) > 1 or self.ENABLE_ORDER_AND_LIMIT_ON_SQL is False
             # multiple statements; execute them, concatenate results and apply
             # ordering and limiting on the final result (in memory)
             df = self._sa_execute(stmts)
-            df = self._df_order(df, is_shuffle, order_by)
-            df = self._df_limit(df, limit_n_rows)
+            df = self._df_order(df, shuffle, order_by)
+            df = self._df_limit(df, limit)
         if do_coerce_dtypes:
             df = coerce_dtypes_by_encoding(df, self.encoding_types)
         _LOG.info(f"read DB data `{self.name}` {df.shape} in {time.time() - t0:.2f}s")
@@ -1275,6 +1290,8 @@ class SqlAlchemyTable(DBTable, abc.ABC):
         assert self.is_output
         _LOG.info(f"write data {df.shape} to table `{self.name}` started")
 
+        df = _normalize_for_sql(df)
+
         # map mostly-data dtypes to SQLAlchemy dtypes
         dtypes = self.dtype_class().from_dtypes_unwrap(self.dtypes)
 
@@ -1303,6 +1320,37 @@ class SqlAlchemyTable(DBTable, abc.ABC):
         return self._sa_execute([stmt]).loc[0, "count_1"]
 
 
+def _normalize_for_sql(df: pd.DataFrame) -> pd.DataFrame:
+    # ensure pd.DataFrame.to_sql(...) works alongside pyarrow backend
+    apply_map = {
+        np.ndarray: lambda x: x.tolist() if isinstance(x, np.ndarray) else x,
+        enum.Enum: lambda x: x.value if isinstance(x, enum.Enum) else x,
+    }
+    astype_map = {
+        decimal.Decimal: FLOAT64,
+        uuid.UUID: STRING,
+    }
+
+    for col in df.columns:
+        series = df[col]
+        if series.empty or series.dropna().empty:
+            continue
+
+        sample = series.dropna().iloc[0]
+
+        for typ, astype_target in astype_map.items():
+            if isinstance(sample, typ):
+                df[col] = series.astype(astype_target)
+                break
+        else:
+            for typ, apply_func in apply_map.items():
+                if isinstance(sample, typ):
+                    df[col] = series.apply(apply_func)
+                    break
+
+    return df
+
+
 def _write_chunk(
     chunk: pd.DataFrame,
     sa_engine_uri: str,
